import 'dart:async';
import 'package:collection/collection.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:_private_core/theme/option_chat.dart';
import 'package:flutter_keyboard_visibility/flutter_keyboard_visibility.dart';
import 'package:_private_chat/utils/utils.dart';
import 'package:get/get.dart' as getx;
import 'package:get/get.dart';
import 'package:_private_core/_private_core.dart';
import '../chat_list/bloc/chat_list_bloc.dart';
import '../options.dart';
import 'bloc/chat_detail_bloc.dart';
import 'package:_private_chat/constant.dart';
import 'message.dart';

ChatDetailBloc get _chatDetailBloc => Get.find<ChatDetailBloc>();

enum ChatConversationType { solo, group }

class ChatDetailParams {
  ChatConversationType conversationType;
  String? conversationId;

  //Bot infos
  bool isBot;

  //Solo type
  String? peerUsername;
  String? peerFirebaseId;

  //Group type
  List<String>? peersUsername;
  List<String>? peersFirebaseId;
  List? peersDetail;

  String? name;
  String? thumbnail;
  int? unread;

  ChatDetailParams(
    this.conversationType, {
    this.conversationId,
    this.peerUsername,
    this.peerFirebaseId,
    this.peersUsername,
    this.peersFirebaseId,
    this.peersDetail,
    this.name,
    this.thumbnail,
    this.unread,
    this.isBot = false,
  });
}

class ChatDetailScreen extends StatefulWidget {
  static double get padding => 16;

  final ChatDetailParams? params;
  final PChatOptionsImpl? options;

  const ChatDetailScreen({
    Key? key,
    this.params,
    this.options,
  }) : super(key: key);

  static String route = '/chat_detail';

  @override
  _ChatDetailScreenState createState() => _ChatDetailScreenState();
}

class _ChatDetailScreenState extends State<ChatDetailScreen>
    with TickerProviderStateMixin {
  late ChatDetailParams params;

  Widget chatDetailBackgroundBuilder(
      Widget child, bool isLoadingList, bool isLoadingDetail) {
    if ((widget.options ?? chatOptions).chatDetailBackgroundBuilder != null) {
      return (widget.options ?? chatOptions).chatDetailBackgroundBuilder!(
          child, isLoadingList, isLoadingDetail);
    }
    return child;
  }

  @override
  void initState() {
    super.initState();
    params = widget.params ?? getx.Get.arguments;
    _chatDetailBloc.params = params;
    _chatDetailBloc.add(ChatDetailInitialEvent(this));
    chatOptions.chatOnInitDetail?.call(this);
  }

  @override
  void dispose() {
    _chatDetailBloc.add(ChatDetailDisposeEvent());
    chatOptions.chatOnDisposeDetail?.call();
    super.dispose(); 
  }

  @override
  Widget build(BuildContext context) {
    Widget child = KeyboardDismissOnTap(
      child: KeyboardVisibilityBuilder(builder: (_, isKeyboardVisible) {
        if (isKeyboardVisible) {
          Timer(const Duration(milliseconds: 500), () {
            if (_chatDetailBloc.scrollController.hasClients) {
              _chatDetailBloc.scrollController.animateTo(
                  _chatDetailBloc.scrollController.position.minScrollExtent,
                  duration: const Duration(milliseconds: 350),
                  curve: Curves.easeIn);
            }
          });
        }
        Widget child = Stack(
          alignment: Alignment.topCenter,
          children: [
            Column(
              children: [
                WidgetProfileChat(
                  params: params,
                  builder: widget.options?.chatDetailProfileBuilder,
                ),
                _buildMessages(),
                Row(
                  children: [
                    const Spacer(),
                    SizedBox(
                      key: _chatDetailBloc.targetAssistantKey,
                      width: 2,
                      height: 2,
                    ),
                  ],
                ),
                WidgetInputChat(
                  params: params,
                  builder: widget.options?.chatDetailInputBuilder,
                ),
              ],
            ),
          ],
        );

        return BlocBuilder<ChatListBloc, ChatListState>(
          buildWhen: (previous, current) =>
              previous.isLoading != current.isLoading,
          bloc: Get.find<ChatListBloc>(),
          builder: (_, stateChatList) {
            return BlocBuilder<ChatDetailBloc, ChatDetailState>(
              bloc: Get.find<ChatDetailBloc>(),
              buildWhen: (previous, current) =>
                  previous.isLoading != current.isLoading,
              builder: (_, stateChatDetail) {
                return chatDetailBackgroundBuilder(
                    child, stateChatList.isLoading, stateChatDetail.isLoading);
              },
            );
          },
        );
      }),
    );
    return child;
  }

  Widget _buildMessages() {
    return Expanded(
      child: BlocBuilder<ChatDetailBloc, ChatDetailState>(
        bloc: Get.find<ChatDetailBloc>(),
        builder: (_, state) {
          List<Message>? messages = state.messages;
          if (messages == null || state.isLoading) return const SizedBox();
          return ListView(
            physics: const BouncingScrollPhysics(),
            padding: EdgeInsets.symmetric(horizontal: ChatDetailScreen.padding),
            controller: _chatDetailBloc.scrollController,
            reverse: true,
            children: _getGroupedMessages(messages),
          );
        },
      ),
    );
  }

  List<Widget> _getGroupedMessages(List<Message> messages) {
    if (chatOptions.chatDetailGroupedMessagesBuilder != null) {
      return chatOptions.chatDetailGroupedMessagesBuilder!(messages);
    }
    List<Widget> groupedMessages =
        List.from(<Widget>[const SizedBox(height: 20)]);
    groupBy<Message, String>(messages, (msg) {
      String datetime =
          DateTime.fromMillisecondsSinceEpoch(msg.timestamp!).formatDateTime();
      int otherMessages = 0;
      List<Message> temp = List.from(messages
          .where((e) =>
              DateTime.fromMillisecondsSinceEpoch(e.timestamp!)
                      .formatDateTime() ==
                  datetime &&
              ((msg.doc?.data() ?? msg.tempDoc) as Map)[kDbFROM] !=
                  ((e.doc?.data() ?? e.tempDoc) as Map)[kDbFROM] &&
              e.timestamp! > msg.timestamp!)
          .toList())
        ..sorted((a, b) => a.timestamp! - b.timestamp!);
      otherMessages = temp.length;
      return '$datetime-${((msg.doc?.data() ?? msg.tempDoc) as Map)[kDbFROM]}-$otherMessages';
    }).forEach((when, actualMessages) {
      groupedMessages.add(Center(
        child: Padding(
          padding: const EdgeInsets.symmetric(vertical: 16),
          child: Text(
            when,
            style: w400TextStyle(color: appColors.text.withOpacity(.6), fontSize: 10),
          ),
        ),
      ));
      groupedMessages
          .add(actualMessages.last.widgetMessageOwner(actualMessages));
    });
    return groupedMessages.reversed.toList();
  }
}

class WidgetProfileChat extends StatelessWidget {
  final ChatDetailParams params;
  final Widget Function(dynamic state, bool isBot)? builder;
  const WidgetProfileChat({
    Key? key,
    required this.params,
    this.builder,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return BlocBuilder<ChatDetailBloc, ChatDetailState>(
      bloc: _chatDetailBloc,
      builder: (_, state) {
        return KeyedSubtree(
          key: ValueKey(state.isLoading),
          child: builder?.call(state, params.isBot) ??
              chatOptions.chatDetailProfileBuilder(state, params.isBot),
        );
      },
    );
  }
}

class WidgetInputChat extends StatelessWidget {
  final ChatDetailParams params;
  final Widget Function(dynamic state, TextEditingController controller,
      FocusNode node, Function() onSend, bool isBot)? builder;

  const WidgetInputChat({
    Key? key,
    required this.params,
    this.builder,
  }) : super(key: key);

  TextEditingController get controller => _chatDetailBloc.inputController;
  FocusNode get inputFocus => _chatDetailBloc.inputFocus;

  void onSend() {
    if (controller.text.trim().isNotEmpty) {
      _chatDetailBloc.add(ChatSendMessageEvent(
          type: MessageType.text, text: controller.text.trim()));
      controller.clear();
    }
  }

  @override
  Widget build(BuildContext context) {
    return BlocBuilder<ChatDetailBloc, ChatDetailState>(
      bloc: _chatDetailBloc,
      builder: (_, state) {
        return builder?.call(
                state, controller, inputFocus, onSend, params.isBot) ??
            chatOptions.chatDetailInputBuilder(
                state, controller, inputFocus, onSend, params.isBot);
      },
    );
  }
}
